/*************************************************************************
Copyright (c) 1992-2007 The University of Tennessee.  All rights reserved.

Contributors:
    * Sergey Bochkanov (ALGLIB project). Translation from FORTRAN to
      pseudocode.

See subroutines comments for additional copyrights.

>>> SOURCE LICENSE >>>
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation (www.fsf.org); either version 2 of the 
License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

A copy of the GNU General Public License is available at
http://www.fsf.org/licensing/licenses

>>> END OF LICENSE >>>
*************************************************************************/

#ifndef _rcond_h
#define _rcond_h

#include "ap.h"
#include "amp.h"
#include "reflections.h"
#include "creflections.h"
#include "hqrnd.h"
#include "matgen.h"
#include "ablasf.h"
#include "ablas.h"
#include "trfac.h"
#include "trlinsolve.h"
#include "safesolve.h"
namespace rcond
{
    
    amp::ampf rmatrixrcond1(ap::template_2d_array< amp::ampf > a,
        int n);
    
    amp::ampf rmatrixrcondinf(ap::template_2d_array< amp::ampf > a,
        int n);
    
    amp::ampf spdmatrixrcond(ap::template_2d_array< amp::ampf > a,
        int n,
        bool isupper);
    
    amp::ampf rmatrixtrrcond1(const ap::template_2d_array< amp::ampf >& a,
        int n,
        bool isupper,
        bool isunit);
    
    amp::ampf rmatrixtrrcondinf(const ap::template_2d_array< amp::ampf >& a,
        int n,
        bool isupper,
        bool isunit);
    
    amp::ampf hpdmatrixrcond(ap::template_2d_array< amp::campf > a,
        int n,
        bool isupper);
    
    amp::ampf cmatrixrcond1(ap::template_2d_array< amp::campf > a,
        int n);
    
    amp::ampf cmatrixrcondinf(ap::template_2d_array< amp::campf > a,
        int n);
    
    amp::ampf rmatrixlurcond1(const ap::template_2d_array< amp::ampf >& lua,
        int n);
    
    amp::ampf rmatrixlurcondinf(const ap::template_2d_array< amp::ampf >& lua,
        int n);
    
    amp::ampf spdmatrixcholeskyrcond(const ap::template_2d_array< amp::ampf >& a,
        int n,
        bool isupper);
    
    amp::ampf hpdmatrixcholeskyrcond(const ap::template_2d_array< amp::campf >& a,
        int n,
        bool isupper);
    
    amp::ampf cmatrixlurcond1(const ap::template_2d_array< amp::campf >& lua,
        int n);
    
    amp::ampf cmatrixlurcondinf(const ap::template_2d_array< amp::campf >& lua,
        int n);
    
    amp::ampf cmatrixtrrcond1(const ap::template_2d_array< amp::campf >& a,
        int n,
        bool isupper,
        bool isunit);
    
    amp::ampf cmatrixtrrcondinf(const ap::template_2d_array< amp::campf >& a,
        int n,
        bool isupper,
        bool isunit);
    
    amp::ampf rcondthreshold();
    
    void rmatrixrcondtrinternal(const ap::template_2d_array< amp::ampf >& a,
        int n,
        bool isupper,
        bool isunit,
        bool onenorm,
        amp::ampf anorm,
        amp::ampf& rc);
    
    void cmatrixrcondtrinternal(const ap::template_2d_array< amp::campf >& a,
        const int& n,
        bool isupper,
        bool isunit,
        bool onenorm,
        amp::ampf anorm,
        amp::ampf& rc);
    
    void spdmatrixrcondcholeskyinternal(const ap::template_2d_array< amp::ampf >& cha,
        int n,
        bool isupper,
        bool isnormprovided,
        amp::ampf anorm,
        amp::ampf& rc);
    
    void hpdmatrixrcondcholeskyinternal(const ap::template_2d_array< amp::campf >& cha,
        int n,
        bool isupper,
        bool isnormprovided,
        amp::ampf anorm,
        amp::ampf& rc);
    
    void rmatrixrcondluinternal(const ap::template_2d_array< amp::ampf >& lua,
        int n,
        bool onenorm,
        bool isanormprovided,
        amp::ampf anorm,
        amp::ampf& rc);
    
    void cmatrixrcondluinternal(const ap::template_2d_array< amp::campf >& lua,
        const int& n,
        bool onenorm,
        bool isanormprovided,
        amp::ampf anorm,
        amp::ampf& rc);
    
    void rmatrixestimatenorm(int n,
        ap::template_1d_array< amp::ampf >& v,
        ap::template_1d_array< amp::ampf >& x,
        ap::template_1d_array< int >& isgn,
        amp::ampf& est,
        int& kase);
    
    void cmatrixestimatenorm(const int& n,
        ap::template_1d_array< amp::campf >& v,
        ap::template_1d_array< amp::campf >& x,
        amp::ampf& est,
        int& kase,
        ap::template_1d_array< int >& isave,
        ap::template_1d_array< amp::ampf >& rsave);
    
    amp::ampf internalcomplexrcondscsum1(const ap::template_1d_array< amp::campf >& x,
        int n);
    
    int internalcomplexrcondicmax1(const ap::template_1d_array< amp::campf >& x,
        int n);
    
    void internalcomplexrcondsaveall(ap::template_1d_array< int >& isave,
        ap::template_1d_array< amp::ampf >& rsave,
        int& i,
        int& iter,
        int& j,
        int& jlast,
        int& jump,
        amp::ampf& absxi,
        amp::ampf& altsgn,
        amp::ampf& estold,
        amp::ampf& temp);
    
    void internalcomplexrcondloadall(ap::template_1d_array< int >& isave,
        ap::template_1d_array< amp::ampf >& rsave,
        int& i,
        int& iter,
        int& j,
        int& jlast,
        int& jump,
        amp::ampf& absxi,
        amp::ampf& altsgn,
        amp::ampf& estold,
        amp::ampf& temp);


    /*************************************************************************
    Estimate of a matrix condition number (1-norm)

    The algorithm calculates a lower bound of the condition number. In this case,
    the algorithm does not return a lower bound of the condition number, but an
    inverse number (to avoid an overflow in case of a singular matrix).

    Input parameters:
        A   -   matrix. Array whose indexes range within [0..N-1, 0..N-1].
        N   -   size of matrix A.

    Result: 1/LowerBound(cond(A))

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf rmatrixrcond1(ap::template_2d_array< amp::ampf > a,
        int n)
    {
        amp::ampf result;
        int i;
        int j;
        amp::ampf v;
        amp::ampf nrm;
        ap::template_1d_array< int > pivots;
        ap::template_1d_array< amp::ampf > t;


        ap::ap_error::make_assertion(n>=1);
        t.setlength(n);
        for(i=0; i<=n-1; i++)
        {
            t(i) = 0;
        }
        for(i=0; i<=n-1; i++)
        {
            for(j=0; j<=n-1; j++)
            {
                t(j) = t(j)+amp::abs(a(i,j));
            }
        }
        nrm = 0;
        for(i=0; i<=n-1; i++)
        {
            nrm = amp::maximum(nrm, t(i));
        }
        trfac::rmatrixlu(a, n, n, pivots);
        rmatrixrcondluinternal(a, n, true, true, nrm, v);
        result = v;
        return result;
    }


    /*************************************************************************
    Estimate of a matrix condition number (infinity-norm).

    The algorithm calculates a lower bound of the condition number. In this case,
    the algorithm does not return a lower bound of the condition number, but an
    inverse number (to avoid an overflow in case of a singular matrix).

    Input parameters:
        A   -   matrix. Array whose indexes range within [0..N-1, 0..N-1].
        N   -   size of matrix A.

    Result: 1/LowerBound(cond(A))

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf rmatrixrcondinf(ap::template_2d_array< amp::ampf > a,
        int n)
    {
        amp::ampf result;
        int i;
        int j;
        amp::ampf v;
        amp::ampf nrm;
        ap::template_1d_array< int > pivots;


        ap::ap_error::make_assertion(n>=1);
        nrm = 0;
        for(i=0; i<=n-1; i++)
        {
            v = 0;
            for(j=0; j<=n-1; j++)
            {
                v = v+amp::abs(a(i,j));
            }
            nrm = amp::maximum(nrm, v);
        }
        trfac::rmatrixlu(a, n, n, pivots);
        rmatrixrcondluinternal(a, n, false, true, nrm, v);
        result = v;
        return result;
    }


    /*************************************************************************
    Condition number estimate of a symmetric positive definite matrix.

    The algorithm calculates a lower bound of the condition number. In this case,
    the algorithm does not return a lower bound of the condition number, but an
    inverse number (to avoid an overflow in case of a singular matrix).

    It should be noted that 1-norm and inf-norm of condition numbers of symmetric
    matrices are equal, so the algorithm doesn't take into account the
    differences between these types of norms.

    Input parameters:
        A       -   symmetric positive definite matrix which is given by its
                    upper or lower triangle depending on the value of
                    IsUpper. Array with elements [0..N-1, 0..N-1].
        N       -   size of matrix A.
        IsUpper -   storage format.

    Result:
        1/LowerBound(cond(A)), if matrix A is positive definite,
       -1, if matrix A is not positive definite, and its condition number
        could not be found by this algorithm.

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf spdmatrixrcond(ap::template_2d_array< amp::ampf > a,
        int n,
        bool isupper)
    {
        amp::ampf result;
        int i;
        int j;
        int j1;
        int j2;
        amp::ampf v;
        amp::ampf nrm;
        ap::template_1d_array< amp::ampf > t;


        t.setlength(n);
        for(i=0; i<=n-1; i++)
        {
            t(i) = 0;
        }
        for(i=0; i<=n-1; i++)
        {
            if( isupper )
            {
                j1 = i;
                j2 = n-1;
            }
            else
            {
                j1 = 0;
                j2 = i;
            }
            for(j=j1; j<=j2; j++)
            {
                if( i==j )
                {
                    t(i) = t(i)+amp::abs(a(i,i));
                }
                else
                {
                    t(i) = t(i)+amp::abs(a(i,j));
                    t(j) = t(j)+amp::abs(a(i,j));
                }
            }
        }
        nrm = 0;
        for(i=0; i<=n-1; i++)
        {
            nrm = amp::maximum(nrm, t(i));
        }
        if( trfac::spdmatrixcholesky(a, n, isupper) )
        {
            spdmatrixrcondcholeskyinternal(a, n, isupper, true, nrm, v);
            result = v;
        }
        else
        {
            result = -1;
        }
        return result;
    }


    /*************************************************************************
    Triangular matrix: estimate of a condition number (1-norm)

    The algorithm calculates a lower bound of the condition number. In this case,
    the algorithm does not return a lower bound of the condition number, but an
    inverse number (to avoid an overflow in case of a singular matrix).

    Input parameters:
        A       -   matrix. Array[0..N-1, 0..N-1].
        N       -   size of A.
        IsUpper -   True, if the matrix is upper triangular.
        IsUnit  -   True, if the matrix has a unit diagonal.

    Result: 1/LowerBound(cond(A))

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf rmatrixtrrcond1(const ap::template_2d_array< amp::ampf >& a,
        int n,
        bool isupper,
        bool isunit)
    {
        amp::ampf result;
        int i;
        int j;
        amp::ampf v;
        amp::ampf nrm;
        ap::template_1d_array< int > pivots;
        ap::template_1d_array< amp::ampf > t;
        int j1;
        int j2;


        ap::ap_error::make_assertion(n>=1);
        t.setlength(n);
        for(i=0; i<=n-1; i++)
        {
            t(i) = 0;
        }
        for(i=0; i<=n-1; i++)
        {
            if( isupper )
            {
                j1 = i+1;
                j2 = n-1;
            }
            else
            {
                j1 = 0;
                j2 = i-1;
            }
            for(j=j1; j<=j2; j++)
            {
                t(j) = t(j)+amp::abs(a(i,j));
            }
            if( isunit )
            {
                t(i) = t(i)+1;
            }
            else
            {
                t(i) = t(i)+amp::abs(a(i,i));
            }
        }
        nrm = 0;
        for(i=0; i<=n-1; i++)
        {
            nrm = amp::maximum(nrm, t(i));
        }
        rmatrixrcondtrinternal(a, n, isupper, isunit, true, nrm, v);
        result = v;
        return result;
    }


    /*************************************************************************
    Triangular matrix: estimate of a matrix condition number (infinity-norm).

    The algorithm calculates a lower bound of the condition number. In this case,
    the algorithm does not return a lower bound of the condition number, but an
    inverse number (to avoid an overflow in case of a singular matrix).

    Input parameters:
        A   -   matrix. Array whose indexes range within [0..N-1, 0..N-1].
        N   -   size of matrix A.
        IsUpper -   True, if the matrix is upper triangular.
        IsUnit  -   True, if the matrix has a unit diagonal.

    Result: 1/LowerBound(cond(A))

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf rmatrixtrrcondinf(const ap::template_2d_array< amp::ampf >& a,
        int n,
        bool isupper,
        bool isunit)
    {
        amp::ampf result;
        int i;
        int j;
        amp::ampf v;
        amp::ampf nrm;
        ap::template_1d_array< int > pivots;
        int j1;
        int j2;


        ap::ap_error::make_assertion(n>=1);
        nrm = 0;
        for(i=0; i<=n-1; i++)
        {
            if( isupper )
            {
                j1 = i+1;
                j2 = n-1;
            }
            else
            {
                j1 = 0;
                j2 = i-1;
            }
            v = 0;
            for(j=j1; j<=j2; j++)
            {
                v = v+amp::abs(a(i,j));
            }
            if( isunit )
            {
                v = v+1;
            }
            else
            {
                v = v+amp::abs(a(i,i));
            }
            nrm = amp::maximum(nrm, v);
        }
        rmatrixrcondtrinternal(a, n, isupper, isunit, false, nrm, v);
        result = v;
        return result;
    }


    /*************************************************************************
    Condition number estimate of a Hermitian positive definite matrix.

    The algorithm calculates a lower bound of the condition number. In this case,
    the algorithm does not return a lower bound of the condition number, but an
    inverse number (to avoid an overflow in case of a singular matrix).

    It should be noted that 1-norm and inf-norm of condition numbers of symmetric
    matrices are equal, so the algorithm doesn't take into account the
    differences between these types of norms.

    Input parameters:
        A       -   Hermitian positive definite matrix which is given by its
                    upper or lower triangle depending on the value of
                    IsUpper. Array with elements [0..N-1, 0..N-1].
        N       -   size of matrix A.
        IsUpper -   storage format.

    Result:
        1/LowerBound(cond(A)), if matrix A is positive definite,
       -1, if matrix A is not positive definite, and its condition number
        could not be found by this algorithm.

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf hpdmatrixrcond(ap::template_2d_array< amp::campf > a,
        int n,
        bool isupper)
    {
        amp::ampf result;
        int i;
        int j;
        int j1;
        int j2;
        amp::ampf v;
        amp::ampf nrm;
        ap::template_1d_array< amp::ampf > t;


        t.setlength(n);
        for(i=0; i<=n-1; i++)
        {
            t(i) = 0;
        }
        for(i=0; i<=n-1; i++)
        {
            if( isupper )
            {
                j1 = i;
                j2 = n-1;
            }
            else
            {
                j1 = 0;
                j2 = i;
            }
            for(j=j1; j<=j2; j++)
            {
                if( i==j )
                {
                    t(i) = t(i)+amp::abscomplex(a(i,i));
                }
                else
                {
                    t(i) = t(i)+amp::abscomplex(a(i,j));
                    t(j) = t(j)+amp::abscomplex(a(i,j));
                }
            }
        }
        nrm = 0;
        for(i=0; i<=n-1; i++)
        {
            nrm = amp::maximum(nrm, t(i));
        }
        if( trfac::hpdmatrixcholesky(a, n, isupper) )
        {
            hpdmatrixrcondcholeskyinternal(a, n, isupper, true, nrm, v);
            result = v;
        }
        else
        {
            result = -1;
        }
        return result;
    }


    /*************************************************************************
    Estimate of a matrix condition number (1-norm)

    The algorithm calculates a lower bound of the condition number. In this case,
    the algorithm does not return a lower bound of the condition number, but an
    inverse number (to avoid an overflow in case of a singular matrix).

    Input parameters:
        A   -   matrix. Array whose indexes range within [0..N-1, 0..N-1].
        N   -   size of matrix A.

    Result: 1/LowerBound(cond(A))

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf cmatrixrcond1(ap::template_2d_array< amp::campf > a,
        int n)
    {
        amp::ampf result;
        int i;
        int j;
        amp::ampf v;
        amp::ampf nrm;
        ap::template_1d_array< int > pivots;
        ap::template_1d_array< amp::ampf > t;


        ap::ap_error::make_assertion(n>=1);
        t.setlength(n);
        for(i=0; i<=n-1; i++)
        {
            t(i) = 0;
        }
        for(i=0; i<=n-1; i++)
        {
            for(j=0; j<=n-1; j++)
            {
                t(j) = t(j)+amp::abscomplex(a(i,j));
            }
        }
        nrm = 0;
        for(i=0; i<=n-1; i++)
        {
            nrm = amp::maximum(nrm, t(i));
        }
        trfac::cmatrixlu(a, n, n, pivots);
        cmatrixrcondluinternal(a, n, true, true, nrm, v);
        result = v;
        return result;
    }


    /*************************************************************************
    Estimate of a matrix condition number (infinity-norm).

    The algorithm calculates a lower bound of the condition number. In this case,
    the algorithm does not return a lower bound of the condition number, but an
    inverse number (to avoid an overflow in case of a singular matrix).

    Input parameters:
        A   -   matrix. Array whose indexes range within [0..N-1, 0..N-1].
        N   -   size of matrix A.

    Result: 1/LowerBound(cond(A))

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf cmatrixrcondinf(ap::template_2d_array< amp::campf > a,
        int n)
    {
        amp::ampf result;
        int i;
        int j;
        amp::ampf v;
        amp::ampf nrm;
        ap::template_1d_array< int > pivots;


        ap::ap_error::make_assertion(n>=1);
        nrm = 0;
        for(i=0; i<=n-1; i++)
        {
            v = 0;
            for(j=0; j<=n-1; j++)
            {
                v = v+amp::abscomplex(a(i,j));
            }
            nrm = amp::maximum(nrm, v);
        }
        trfac::cmatrixlu(a, n, n, pivots);
        cmatrixrcondluinternal(a, n, false, true, nrm, v);
        result = v;
        return result;
    }


    /*************************************************************************
    Estimate of the condition number of a matrix given by its LU decomposition (1-norm)

    The algorithm calculates a lower bound of the condition number. In this case,
    the algorithm does not return a lower bound of the condition number, but an
    inverse number (to avoid an overflow in case of a singular matrix).

    Input parameters:
        LUA         -   LU decomposition of a matrix in compact form. Output of
                        the RMatrixLU subroutine.
        N           -   size of matrix A.

    Result: 1/LowerBound(cond(A))

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf rmatrixlurcond1(const ap::template_2d_array< amp::ampf >& lua,
        int n)
    {
        amp::ampf result;
        amp::ampf v;


        rmatrixrcondluinternal(lua, n, true, false, amp::ampf(0), v);
        result = v;
        return result;
    }


    /*************************************************************************
    Estimate of the condition number of a matrix given by its LU decomposition
    (infinity norm).

    The algorithm calculates a lower bound of the condition number. In this case,
    the algorithm does not return a lower bound of the condition number, but an
    inverse number (to avoid an overflow in case of a singular matrix).

    Input parameters:
        LUA     -   LU decomposition of a matrix in compact form. Output of
                    the RMatrixLU subroutine.
        N       -   size of matrix A.

    Result: 1/LowerBound(cond(A))

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf rmatrixlurcondinf(const ap::template_2d_array< amp::ampf >& lua,
        int n)
    {
        amp::ampf result;
        amp::ampf v;


        rmatrixrcondluinternal(lua, n, false, false, amp::ampf(0), v);
        result = v;
        return result;
    }


    /*************************************************************************
    Condition number estimate of a symmetric positive definite matrix given by
    Cholesky decomposition.

    The algorithm calculates a lower bound of the condition number. In this
    case, the algorithm does not return a lower bound of the condition number,
    but an inverse number (to avoid an overflow in case of a singular matrix).

    It should be noted that 1-norm and inf-norm condition numbers of symmetric
    matrices are equal, so the algorithm doesn't take into account the
    differences between these types of norms.

    Input parameters:
        CD  - Cholesky decomposition of matrix A,
              output of SMatrixCholesky subroutine.
        N   - size of matrix A.

    Result: 1/LowerBound(cond(A))

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf spdmatrixcholeskyrcond(const ap::template_2d_array< amp::ampf >& a,
        int n,
        bool isupper)
    {
        amp::ampf result;
        amp::ampf v;


        spdmatrixrcondcholeskyinternal(a, n, isupper, false, amp::ampf(0), v);
        result = v;
        return result;
    }


    /*************************************************************************
    Condition number estimate of a Hermitian positive definite matrix given by
    Cholesky decomposition.

    The algorithm calculates a lower bound of the condition number. In this
    case, the algorithm does not return a lower bound of the condition number,
    but an inverse number (to avoid an overflow in case of a singular matrix).

    It should be noted that 1-norm and inf-norm condition numbers of symmetric
    matrices are equal, so the algorithm doesn't take into account the
    differences between these types of norms.

    Input parameters:
        CD  - Cholesky decomposition of matrix A,
              output of SMatrixCholesky subroutine.
        N   - size of matrix A.

    Result: 1/LowerBound(cond(A))

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf hpdmatrixcholeskyrcond(const ap::template_2d_array< amp::campf >& a,
        int n,
        bool isupper)
    {
        amp::ampf result;
        amp::ampf v;


        hpdmatrixrcondcholeskyinternal(a, n, isupper, false, amp::ampf(0), v);
        result = v;
        return result;
    }


    /*************************************************************************
    Estimate of the condition number of a matrix given by its LU decomposition (1-norm)

    The algorithm calculates a lower bound of the condition number. In this case,
    the algorithm does not return a lower bound of the condition number, but an
    inverse number (to avoid an overflow in case of a singular matrix).

    Input parameters:
        LUA         -   LU decomposition of a matrix in compact form. Output of
                        the CMatrixLU subroutine.
        N           -   size of matrix A.

    Result: 1/LowerBound(cond(A))

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf cmatrixlurcond1(const ap::template_2d_array< amp::campf >& lua,
        int n)
    {
        amp::ampf result;
        amp::ampf v;


        ap::ap_error::make_assertion(n>=1);
        cmatrixrcondluinternal(lua, n, true, false, amp::ampf("0.0"), v);
        result = v;
        return result;
    }


    /*************************************************************************
    Estimate of the condition number of a matrix given by its LU decomposition
    (infinity norm).

    The algorithm calculates a lower bound of the condition number. In this case,
    the algorithm does not return a lower bound of the condition number, but an
    inverse number (to avoid an overflow in case of a singular matrix).

    Input parameters:
        LUA     -   LU decomposition of a matrix in compact form. Output of
                    the CMatrixLU subroutine.
        N       -   size of matrix A.

    Result: 1/LowerBound(cond(A))

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf cmatrixlurcondinf(const ap::template_2d_array< amp::campf >& lua,
        int n)
    {
        amp::ampf result;
        amp::ampf v;


        ap::ap_error::make_assertion(n>=1);
        cmatrixrcondluinternal(lua, n, false, false, amp::ampf("0.0"), v);
        result = v;
        return result;
    }


    /*************************************************************************
    Triangular matrix: estimate of a condition number (1-norm)

    The algorithm calculates a lower bound of the condition number. In this case,
    the algorithm does not return a lower bound of the condition number, but an
    inverse number (to avoid an overflow in case of a singular matrix).

    Input parameters:
        A       -   matrix. Array[0..N-1, 0..N-1].
        N       -   size of A.
        IsUpper -   True, if the matrix is upper triangular.
        IsUnit  -   True, if the matrix has a unit diagonal.

    Result: 1/LowerBound(cond(A))

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf cmatrixtrrcond1(const ap::template_2d_array< amp::campf >& a,
        int n,
        bool isupper,
        bool isunit)
    {
        amp::ampf result;
        int i;
        int j;
        amp::ampf v;
        amp::ampf nrm;
        ap::template_1d_array< int > pivots;
        ap::template_1d_array< amp::ampf > t;
        int j1;
        int j2;


        ap::ap_error::make_assertion(n>=1);
        t.setlength(n);
        for(i=0; i<=n-1; i++)
        {
            t(i) = 0;
        }
        for(i=0; i<=n-1; i++)
        {
            if( isupper )
            {
                j1 = i+1;
                j2 = n-1;
            }
            else
            {
                j1 = 0;
                j2 = i-1;
            }
            for(j=j1; j<=j2; j++)
            {
                t(j) = t(j)+amp::abscomplex(a(i,j));
            }
            if( isunit )
            {
                t(i) = t(i)+1;
            }
            else
            {
                t(i) = t(i)+amp::abscomplex(a(i,i));
            }
        }
        nrm = 0;
        for(i=0; i<=n-1; i++)
        {
            nrm = amp::maximum(nrm, t(i));
        }
        cmatrixrcondtrinternal(a, n, isupper, isunit, true, nrm, v);
        result = v;
        return result;
    }


    /*************************************************************************
    Triangular matrix: estimate of a matrix condition number (infinity-norm).

    The algorithm calculates a lower bound of the condition number. In this case,
    the algorithm does not return a lower bound of the condition number, but an
    inverse number (to avoid an overflow in case of a singular matrix).

    Input parameters:
        A   -   matrix. Array whose indexes range within [0..N-1, 0..N-1].
        N   -   size of matrix A.
        IsUpper -   True, if the matrix is upper triangular.
        IsUnit  -   True, if the matrix has a unit diagonal.

    Result: 1/LowerBound(cond(A))

    NOTE:
        if k(A) is very large, then matrix is  assumed  degenerate,  k(A)=INF,
        0.0 is returned in such cases.
    *************************************************************************/
    
    amp::ampf cmatrixtrrcondinf(const ap::template_2d_array< amp::campf >& a,
        int n,
        bool isupper,
        bool isunit)
    {
        amp::ampf result;
        int i;
        int j;
        amp::ampf v;
        amp::ampf nrm;
        ap::template_1d_array< int > pivots;
        int j1;
        int j2;


        ap::ap_error::make_assertion(n>=1);
        nrm = 0;
        for(i=0; i<=n-1; i++)
        {
            if( isupper )
            {
                j1 = i+1;
                j2 = n-1;
            }
            else
            {
                j1 = 0;
                j2 = i-1;
            }
            v = 0;
            for(j=j1; j<=j2; j++)
            {
                v = v+amp::abscomplex(a(i,j));
            }
            if( isunit )
            {
                v = v+1;
            }
            else
            {
                v = v+amp::abscomplex(a(i,i));
            }
            nrm = amp::maximum(nrm, v);
        }
        cmatrixrcondtrinternal(a, n, isupper, isunit, false, nrm, v);
        result = v;
        return result;
    }


    /*************************************************************************
    Threshold for rcond: matrices with condition number beyond this  threshold
    are considered singular.

    Threshold must be far enough from underflow, at least Sqr(Threshold)  must
    be greater than underflow.
    *************************************************************************/
    
    amp::ampf rcondthreshold()
    {
        amp::ampf result;


        result = amp::sqrt(amp::sqrt(amp::ampf::getAlgoPascalMinNumber()));
        return result;
    }


    /*************************************************************************
    Internal subroutine for condition number estimation

      -- LAPACK routine (version 3.0) --
         Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
         Courant Institute, Argonne National Lab, and Rice University
         February 29, 1992
    *************************************************************************/
    
    void rmatrixrcondtrinternal(const ap::template_2d_array< amp::ampf >& a,
        int n,
        bool isupper,
        bool isunit,
        bool onenorm,
        amp::ampf anorm,
        amp::ampf& rc)
    {
        ap::template_1d_array< amp::ampf > ex;
        ap::template_1d_array< amp::ampf > ev;
        ap::template_1d_array< int > iwork;
        ap::template_1d_array< amp::ampf > tmp;
        amp::ampf v;
        int i;
        int j;
        int kase;
        int kase1;
        int j1;
        int j2;
        amp::ampf ainvnm;
        amp::ampf maxgrowth;
        amp::ampf s;
        bool mupper;
        bool mtrans;
        bool munit;


        
        //
        // RC=0 if something happens
        //
        rc = 0;
        
        //
        // init
        //
        if( onenorm )
        {
            kase1 = 1;
        }
        else
        {
            kase1 = 2;
        }
        mupper = true;
        mtrans = true;
        munit = true;
        iwork.setlength(n+1);
        tmp.setlength(n);
        
        //
        // prepare parameters for triangular solver
        //
        maxgrowth = 1/rcondthreshold();
        s = 0;
        for(i=0; i<=n-1; i++)
        {
            if( isupper )
            {
                j1 = i+1;
                j2 = n-1;
            }
            else
            {
                j1 = 0;
                j2 = i-1;
            }
            for(j=j1; j<=j2; j++)
            {
                s = amp::maximum(s, amp::abs(a(i,j)));
            }
            if( isunit )
            {
                s = amp::maximum(s, amp::ampf(1));
            }
            else
            {
                s = amp::maximum(s, amp::abs(a(i,i)));
            }
        }
        if( s==0 )
        {
            s = 1;
        }
        s = 1/s;
        
        //
        // Scale according to S
        //
        anorm = anorm*s;
        
        //
        // Quick return if possible
        // We assume that ANORM<>0 after this block
        //
        if( anorm==0 )
        {
            return;
        }
        if( n==1 )
        {
            rc = 1;
            return;
        }
        
        //
        // Estimate the norm of inv(A).
        //
        ainvnm = 0;
        kase = 0;
        while( true )
        {
            rmatrixestimatenorm(n, ev, ex, iwork, ainvnm, kase);
            if( kase==0 )
            {
                break;
            }
            
            //
            // from 1-based array to 0-based
            //
            for(i=0; i<=n-1; i++)
            {
                ex(i) = ex(i+1);
            }
            
            //
            // multiply by inv(A) or inv(A')
            //
            if( kase==kase1 )
            {
                
                //
                // multiply by inv(A)
                //
                if( !safesolve::rmatrixscaledtrsafesolve(a, s, n, ex, isupper, 0, isunit, maxgrowth) )
                {
                    return;
                }
            }
            else
            {
                
                //
                // multiply by inv(A')
                //
                if( !safesolve::rmatrixscaledtrsafesolve(a, s, n, ex, isupper, 1, isunit, maxgrowth) )
                {
                    return;
                }
            }
            
            //
            // from 0-based array to 1-based
            //
            for(i=n-1; i>=0; i--)
            {
                ex(i+1) = ex(i);
            }
        }
        
        //
        // Compute the estimate of the reciprocal condition number.
        //
        if( ainvnm!=0 )
        {
            rc = 1/ainvnm;
            rc = rc/anorm;
            if( rc<rcondthreshold() )
            {
                rc = 0;
            }
        }
    }


    /*************************************************************************
    Condition number estimation

      -- LAPACK routine (version 3.0) --
         Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
         Courant Institute, Argonne National Lab, and Rice University
         March 31, 1993
    *************************************************************************/
    
    void cmatrixrcondtrinternal(const ap::template_2d_array< amp::campf >& a,
        const int& n,
        bool isupper,
        bool isunit,
        bool onenorm,
        amp::ampf anorm,
        amp::ampf& rc)
    {
        ap::template_1d_array< amp::campf > ex;
        ap::template_1d_array< amp::campf > cwork2;
        ap::template_1d_array< amp::campf > cwork3;
        ap::template_1d_array< amp::campf > cwork4;
        ap::template_1d_array< int > isave;
        ap::template_1d_array< amp::ampf > rsave;
        int kase;
        int kase1;
        amp::ampf ainvnm;
        amp::campf v;
        int i;
        int j;
        int j1;
        int j2;
        amp::ampf s;
        amp::ampf maxgrowth;


        
        //
        // RC=0 if something happens
        //
        rc = 0;
        
        //
        // init
        //
        if( n<=0 )
        {
            return;
        }
        if( n==0 )
        {
            rc = 1;
            return;
        }
        cwork2.setlength(n+1);
        
        //
        // prepare parameters for triangular solver
        //
        maxgrowth = 1/rcondthreshold();
        s = 0;
        for(i=0; i<=n-1; i++)
        {
            if( isupper )
            {
                j1 = i+1;
                j2 = n-1;
            }
            else
            {
                j1 = 0;
                j2 = i-1;
            }
            for(j=j1; j<=j2; j++)
            {
                s = amp::maximum(s, amp::abscomplex(a(i,j)));
            }
            if( isunit )
            {
                s = amp::maximum(s, amp::ampf(1));
            }
            else
            {
                s = amp::maximum(s, amp::abscomplex(a(i,i)));
            }
        }
        if( s==0 )
        {
            s = 1;
        }
        s = 1/s;
        
        //
        // Scale according to S
        //
        anorm = anorm*s;
        
        //
        // Quick return if possible
        //
        if( anorm==0 )
        {
            return;
        }
        
        //
        // Estimate the norm of inv(A).
        //
        ainvnm = 0;
        if( onenorm )
        {
            kase1 = 1;
        }
        else
        {
            kase1 = 2;
        }
        kase = 0;
        while( true )
        {
            cmatrixestimatenorm(n, cwork4, ex, ainvnm, kase, isave, rsave);
            if( kase==0 )
            {
                break;
            }
            
            //
            // From 1-based to 0-based
            //
            for(i=0; i<=n-1; i++)
            {
                ex(i) = ex(i+1);
            }
            
            //
            // multiply by inv(A) or inv(A')
            //
            if( kase==kase1 )
            {
                
                //
                // multiply by inv(A)
                //
                if( !safesolve::cmatrixscaledtrsafesolve(a, s, n, ex, isupper, 0, isunit, maxgrowth) )
                {
                    return;
                }
            }
            else
            {
                
                //
                // multiply by inv(A')
                //
                if( !safesolve::cmatrixscaledtrsafesolve(a, s, n, ex, isupper, 2, isunit, maxgrowth) )
                {
                    return;
                }
            }
            
            //
            // from 0-based to 1-based
            //
            for(i=n-1; i>=0; i--)
            {
                ex(i+1) = ex(i);
            }
        }
        
        //
        // Compute the estimate of the reciprocal condition number.
        //
        if( ainvnm!=0 )
        {
            rc = 1/ainvnm;
            rc = rc/anorm;
            if( rc<rcondthreshold() )
            {
                rc = 0;
            }
        }
    }


    /*************************************************************************
    Internal subroutine for condition number estimation

      -- LAPACK routine (version 3.0) --
         Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
         Courant Institute, Argonne National Lab, and Rice University
         February 29, 1992
    *************************************************************************/
    
    void spdmatrixrcondcholeskyinternal(const ap::template_2d_array< amp::ampf >& cha,
        int n,
        bool isupper,
        bool isnormprovided,
        amp::ampf anorm,
        amp::ampf& rc)
    {
        int i;
        int j;
        int kase;
        amp::ampf ainvnm;
        ap::template_1d_array< amp::ampf > ex;
        ap::template_1d_array< amp::ampf > ev;
        ap::template_1d_array< amp::ampf > tmp;
        ap::template_1d_array< int > iwork;
        amp::ampf sa;
        amp::ampf v;
        amp::ampf maxgrowth;


        ap::ap_error::make_assertion(n>=1);
        tmp.setlength(n);
        
        //
        // RC=0 if something happens
        //
        rc = 0;
        
        //
        // prepare parameters for triangular solver
        //
        maxgrowth = 1/rcondthreshold();
        sa = 0;
        if( isupper )
        {
            for(i=0; i<=n-1; i++)
            {
                for(j=i; j<=n-1; j++)
                {
                    sa = amp::maximum(sa, amp::abscomplex(cha(i,j)));
                }
            }
        }
        else
        {
            for(i=0; i<=n-1; i++)
            {
                for(j=0; j<=i; j++)
                {
                    sa = amp::maximum(sa, amp::abscomplex(cha(i,j)));
                }
            }
        }
        if( sa==0 )
        {
            sa = 1;
        }
        sa = 1/sa;
        
        //
        // Estimate the norm of A.
        //
        if( !isnormprovided )
        {
            kase = 0;
            anorm = 0;
            while( true )
            {
                rmatrixestimatenorm(n, ev, ex, iwork, anorm, kase);
                if( kase==0 )
                {
                    break;
                }
                if( isupper )
                {
                    
                    //
                    // Multiply by U
                    //
                    for(i=1; i<=n; i++)
                    {
                        v = amp::vdotproduct(cha.getrow(i-1, i-1, n-1), ex.getvector(i, n));
                        ex(i) = v;
                    }
                    amp::vmul(ex.getvector(1, n), sa);
                    
                    //
                    // Multiply by U'
                    //
                    for(i=0; i<=n-1; i++)
                    {
                        tmp(i) = 0;
                    }
                    for(i=0; i<=n-1; i++)
                    {
                        v = ex(i+1);
                        amp::vadd(tmp.getvector(i, n-1), cha.getrow(i, i, n-1), v);
                    }
                    amp::vmove(ex.getvector(1, n), tmp.getvector(0, n-1));
                    amp::vmul(ex.getvector(1, n), sa);
                }
                else
                {
                    
                    //
                    // Multiply by L'
                    //
                    for(i=0; i<=n-1; i++)
                    {
                        tmp(i) = 0;
                    }
                    for(i=0; i<=n-1; i++)
                    {
                        v = ex(i+1);
                        amp::vadd(tmp.getvector(0, i), cha.getrow(i, 0, i), v);
                    }
                    amp::vmove(ex.getvector(1, n), tmp.getvector(0, n-1));
                    amp::vmul(ex.getvector(1, n), sa);
                    
                    //
                    // Multiply by L
                    //
                    for(i=n; i>=1; i--)
                    {
                        v = amp::vdotproduct(cha.getrow(i-1, 0, i-1), ex.getvector(1, i));
                        ex(i) = v;
                    }
                    amp::vmul(ex.getvector(1, n), sa);
                }
            }
        }
        
        //
        // Quick return if possible
        //
        if( anorm==0 )
        {
            return;
        }
        if( n==1 )
        {
            rc = 1;
            return;
        }
        
        //
        // Estimate the 1-norm of inv(A).
        //
        kase = 0;
        while( true )
        {
            rmatrixestimatenorm(n, ev, ex, iwork, ainvnm, kase);
            if( kase==0 )
            {
                break;
            }
            for(i=0; i<=n-1; i++)
            {
                ex(i) = ex(i+1);
            }
            if( isupper )
            {
                
                //
                // Multiply by inv(U').
                //
                if( !safesolve::rmatrixscaledtrsafesolve(cha, sa, n, ex, isupper, 1, false, maxgrowth) )
                {
                    return;
                }
                
                //
                // Multiply by inv(U).
                //
                if( !safesolve::rmatrixscaledtrsafesolve(cha, sa, n, ex, isupper, 0, false, maxgrowth) )
                {
                    return;
                }
            }
            else
            {
                
                //
                // Multiply by inv(L).
                //
                if( !safesolve::rmatrixscaledtrsafesolve(cha, sa, n, ex, isupper, 0, false, maxgrowth) )
                {
                    return;
                }
                
                //
                // Multiply by inv(L').
                //
                if( !safesolve::rmatrixscaledtrsafesolve(cha, sa, n, ex, isupper, 1, false, maxgrowth) )
                {
                    return;
                }
            }
            for(i=n-1; i>=0; i--)
            {
                ex(i+1) = ex(i);
            }
        }
        
        //
        // Compute the estimate of the reciprocal condition number.
        //
        if( ainvnm!=0 )
        {
            v = 1/ainvnm;
            rc = v/anorm;
            if( rc<rcondthreshold() )
            {
                rc = 0;
            }
        }
    }


    /*************************************************************************
    Internal subroutine for condition number estimation

      -- LAPACK routine (version 3.0) --
         Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
         Courant Institute, Argonne National Lab, and Rice University
         February 29, 1992
    *************************************************************************/
    
    void hpdmatrixrcondcholeskyinternal(const ap::template_2d_array< amp::campf >& cha,
        int n,
        bool isupper,
        bool isnormprovided,
        amp::ampf anorm,
        amp::ampf& rc)
    {
        ap::template_1d_array< int > isave;
        ap::template_1d_array< amp::ampf > rsave;
        ap::template_1d_array< amp::campf > ex;
        ap::template_1d_array< amp::campf > ev;
        ap::template_1d_array< amp::campf > tmp;
        int kase;
        amp::ampf ainvnm;
        amp::campf v;
        int i;
        int j;
        amp::ampf sa;
        amp::ampf maxgrowth;
        int i_;
        int i1_;


        ap::ap_error::make_assertion(n>=1);
        tmp.setlength(n);
        
        //
        // RC=0 if something happens
        //
        rc = 0;
        
        //
        // prepare parameters for triangular solver
        //
        maxgrowth = 1/rcondthreshold();
        sa = 0;
        if( isupper )
        {
            for(i=0; i<=n-1; i++)
            {
                for(j=i; j<=n-1; j++)
                {
                    sa = amp::maximum(sa, amp::abscomplex(cha(i,j)));
                }
            }
        }
        else
        {
            for(i=0; i<=n-1; i++)
            {
                for(j=0; j<=i; j++)
                {
                    sa = amp::maximum(sa, amp::abscomplex(cha(i,j)));
                }
            }
        }
        if( sa==0 )
        {
            sa = 1;
        }
        sa = 1/sa;
        
        //
        // Estimate the norm of A
        //
        if( !isnormprovided )
        {
            anorm = 0;
            kase = 0;
            while( true )
            {
                cmatrixestimatenorm(n, ev, ex, anorm, kase, isave, rsave);
                if( kase==0 )
                {
                    break;
                }
                if( isupper )
                {
                    
                    //
                    // Multiply by U
                    //
                    for(i=1; i<=n; i++)
                    {
                        i1_ = (i)-(i-1);
                        v = 0.0;
                        for(i_=i-1; i_<=n-1;i_++)
                        {
                            v += cha(i-1,i_)*ex(i_+i1_);
                        }
                        ex(i) = v;
                    }
                    for(i_=1; i_<=n;i_++)
                    {
                        ex(i_) = sa*ex(i_);
                    }
                    
                    //
                    // Multiply by U'
                    //
                    for(i=0; i<=n-1; i++)
                    {
                        tmp(i) = 0;
                    }
                    for(i=0; i<=n-1; i++)
                    {
                        v = ex(i+1);
                        for(i_=i; i_<=n-1;i_++)
                        {
                            tmp(i_) = tmp(i_) + v*amp::conj(cha(i,i_));
                        }
                    }
                    i1_ = (0) - (1);
                    for(i_=1; i_<=n;i_++)
                    {
                        ex(i_) = tmp(i_+i1_);
                    }
                    for(i_=1; i_<=n;i_++)
                    {
                        ex(i_) = sa*ex(i_);
                    }
                }
                else
                {
                    
                    //
                    // Multiply by L'
                    //
                    for(i=0; i<=n-1; i++)
                    {
                        tmp(i) = 0;
                    }
                    for(i=0; i<=n-1; i++)
                    {
                        v = ex(i+1);
                        for(i_=0; i_<=i;i_++)
                        {
                            tmp(i_) = tmp(i_) + v*amp::conj(cha(i,i_));
                        }
                    }
                    i1_ = (0) - (1);
                    for(i_=1; i_<=n;i_++)
                    {
                        ex(i_) = tmp(i_+i1_);
                    }
                    for(i_=1; i_<=n;i_++)
                    {
                        ex(i_) = sa*ex(i_);
                    }
                    
                    //
                    // Multiply by L
                    //
                    for(i=n; i>=1; i--)
                    {
                        i1_ = (1)-(0);
                        v = 0.0;
                        for(i_=0; i_<=i-1;i_++)
                        {
                            v += cha(i-1,i_)*ex(i_+i1_);
                        }
                        ex(i) = v;
                    }
                    for(i_=1; i_<=n;i_++)
                    {
                        ex(i_) = sa*ex(i_);
                    }
                }
            }
        }
        
        //
        // Quick return if possible
        // After this block we assume that ANORM<>0
        //
        if( anorm==0 )
        {
            return;
        }
        if( n==1 )
        {
            rc = 1;
            return;
        }
        
        //
        // Estimate the norm of inv(A).
        //
        ainvnm = 0;
        kase = 0;
        while( true )
        {
            cmatrixestimatenorm(n, ev, ex, ainvnm, kase, isave, rsave);
            if( kase==0 )
            {
                break;
            }
            for(i=0; i<=n-1; i++)
            {
                ex(i) = ex(i+1);
            }
            if( isupper )
            {
                
                //
                // Multiply by inv(U').
                //
                if( !safesolve::cmatrixscaledtrsafesolve(cha, sa, n, ex, isupper, 2, false, maxgrowth) )
                {
                    return;
                }
                
                //
                // Multiply by inv(U).
                //
                if( !safesolve::cmatrixscaledtrsafesolve(cha, sa, n, ex, isupper, 0, false, maxgrowth) )
                {
                    return;
                }
            }
            else
            {
                
                //
                // Multiply by inv(L).
                //
                if( !safesolve::cmatrixscaledtrsafesolve(cha, sa, n, ex, isupper, 0, false, maxgrowth) )
                {
                    return;
                }
                
                //
                // Multiply by inv(L').
                //
                if( !safesolve::cmatrixscaledtrsafesolve(cha, sa, n, ex, isupper, 2, false, maxgrowth) )
                {
                    return;
                }
            }
            for(i=n-1; i>=0; i--)
            {
                ex(i+1) = ex(i);
            }
        }
        
        //
        // Compute the estimate of the reciprocal condition number.
        //
        if( ainvnm!=0 )
        {
            rc = 1/ainvnm;
            rc = rc/anorm;
            if( rc<rcondthreshold() )
            {
                rc = 0;
            }
        }
    }


    /*************************************************************************
    Internal subroutine for condition number estimation

      -- LAPACK routine (version 3.0) --
         Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
         Courant Institute, Argonne National Lab, and Rice University
         February 29, 1992
    *************************************************************************/
    
    void rmatrixrcondluinternal(const ap::template_2d_array< amp::ampf >& lua,
        int n,
        bool onenorm,
        bool isanormprovided,
        amp::ampf anorm,
        amp::ampf& rc)
    {
        ap::template_1d_array< amp::ampf > ex;
        ap::template_1d_array< amp::ampf > ev;
        ap::template_1d_array< int > iwork;
        ap::template_1d_array< amp::ampf > tmp;
        amp::ampf v;
        int i;
        int j;
        int kase;
        int kase1;
        amp::ampf ainvnm;
        amp::ampf maxgrowth;
        amp::ampf su;
        amp::ampf sl;
        bool mupper;
        bool mtrans;
        bool munit;


        
        //
        // RC=0 if something happens
        //
        rc = 0;
        
        //
        // init
        //
        if( onenorm )
        {
            kase1 = 1;
        }
        else
        {
            kase1 = 2;
        }
        mupper = true;
        mtrans = true;
        munit = true;
        iwork.setlength(n+1);
        tmp.setlength(n);
        
        //
        // prepare parameters for triangular solver
        //
        maxgrowth = 1/rcondthreshold();
        su = 0;
        sl = 1;
        for(i=0; i<=n-1; i++)
        {
            for(j=0; j<=i-1; j++)
            {
                sl = amp::maximum(sl, amp::abs(lua(i,j)));
            }
            for(j=i; j<=n-1; j++)
            {
                su = amp::maximum(su, amp::abs(lua(i,j)));
            }
        }
        if( su==0 )
        {
            su = 1;
        }
        su = 1/su;
        sl = 1/sl;
        
        //
        // Estimate the norm of A.
        //
        if( !isanormprovided )
        {
            kase = 0;
            anorm = 0;
            while( true )
            {
                rmatrixestimatenorm(n, ev, ex, iwork, anorm, kase);
                if( kase==0 )
                {
                    break;
                }
                if( kase==kase1 )
                {
                    
                    //
                    // Multiply by U
                    //
                    for(i=1; i<=n; i++)
                    {
                        v = amp::vdotproduct(lua.getrow(i-1, i-1, n-1), ex.getvector(i, n));
                        ex(i) = v;
                    }
                    
                    //
                    // Multiply by L
                    //
                    for(i=n; i>=1; i--)
                    {
                        if( i>1 )
                        {
                            v = amp::vdotproduct(lua.getrow(i-1, 0, i-2), ex.getvector(1, i-1));
                        }
                        else
                        {
                            v = 0;
                        }
                        ex(i) = ex(i)+v;
                    }
                }
                else
                {
                    
                    //
                    // Multiply by L'
                    //
                    for(i=0; i<=n-1; i++)
                    {
                        tmp(i) = 0;
                    }
                    for(i=0; i<=n-1; i++)
                    {
                        v = ex(i+1);
                        if( i>=1 )
                        {
                            amp::vadd(tmp.getvector(0, i-1), lua.getrow(i, 0, i-1), v);
                        }
                        tmp(i) = tmp(i)+v;
                    }
                    amp::vmove(ex.getvector(1, n), tmp.getvector(0, n-1));
                    
                    //
                    // Multiply by U'
                    //
                    for(i=0; i<=n-1; i++)
                    {
                        tmp(i) = 0;
                    }
                    for(i=0; i<=n-1; i++)
                    {
                        v = ex(i+1);
                        amp::vadd(tmp.getvector(i, n-1), lua.getrow(i, i, n-1), v);
                    }
                    amp::vmove(ex.getvector(1, n), tmp.getvector(0, n-1));
                }
            }
        }
        
        //
        // Scale according to SU/SL
        //
        anorm = anorm*su*sl;
        
        //
        // Quick return if possible
        // We assume that ANORM<>0 after this block
        //
        if( anorm==0 )
        {
            return;
        }
        if( n==1 )
        {
            rc = 1;
            return;
        }
        
        //
        // Estimate the norm of inv(A).
        //
        ainvnm = 0;
        kase = 0;
        while( true )
        {
            rmatrixestimatenorm(n, ev, ex, iwork, ainvnm, kase);
            if( kase==0 )
            {
                break;
            }
            
            //
            // from 1-based array to 0-based
            //
            for(i=0; i<=n-1; i++)
            {
                ex(i) = ex(i+1);
            }
            
            //
            // multiply by inv(A) or inv(A')
            //
            if( kase==kase1 )
            {
                
                //
                // Multiply by inv(L).
                //
                if( !safesolve::rmatrixscaledtrsafesolve(lua, sl, n, ex, !mupper, 0, munit, maxgrowth) )
                {
                    return;
                }
                
                //
                // Multiply by inv(U).
                //
                if( !safesolve::rmatrixscaledtrsafesolve(lua, su, n, ex, mupper, 0, !munit, maxgrowth) )
                {
                    return;
                }
            }
            else
            {
                
                //
                // Multiply by inv(U').
                //
                if( !safesolve::rmatrixscaledtrsafesolve(lua, su, n, ex, mupper, 1, !munit, maxgrowth) )
                {
                    return;
                }
                
                //
                // Multiply by inv(L').
                //
                if( !safesolve::rmatrixscaledtrsafesolve(lua, sl, n, ex, !mupper, 1, munit, maxgrowth) )
                {
                    return;
                }
            }
            
            //
            // from 0-based array to 1-based
            //
            for(i=n-1; i>=0; i--)
            {
                ex(i+1) = ex(i);
            }
        }
        
        //
        // Compute the estimate of the reciprocal condition number.
        //
        if( ainvnm!=0 )
        {
            rc = 1/ainvnm;
            rc = rc/anorm;
            if( rc<rcondthreshold() )
            {
                rc = 0;
            }
        }
    }


    /*************************************************************************
    Condition number estimation

      -- LAPACK routine (version 3.0) --
         Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
         Courant Institute, Argonne National Lab, and Rice University
         March 31, 1993
    *************************************************************************/
    
    void cmatrixrcondluinternal(const ap::template_2d_array< amp::campf >& lua,
        const int& n,
        bool onenorm,
        bool isanormprovided,
        amp::ampf anorm,
        amp::ampf& rc)
    {
        ap::template_1d_array< amp::campf > ex;
        ap::template_1d_array< amp::campf > cwork2;
        ap::template_1d_array< amp::campf > cwork3;
        ap::template_1d_array< amp::campf > cwork4;
        ap::template_1d_array< int > isave;
        ap::template_1d_array< amp::ampf > rsave;
        int kase;
        int kase1;
        amp::ampf ainvnm;
        amp::campf v;
        int i;
        int j;
        amp::ampf su;
        amp::ampf sl;
        amp::ampf maxgrowth;
        int i_;
        int i1_;


        if( n<=0 )
        {
            return;
        }
        cwork2.setlength(n+1);
        rc = 0;
        if( n==0 )
        {
            rc = 1;
            return;
        }
        
        //
        // prepare parameters for triangular solver
        //
        maxgrowth = 1/rcondthreshold();
        su = 0;
        sl = 1;
        for(i=0; i<=n-1; i++)
        {
            for(j=0; j<=i-1; j++)
            {
                sl = amp::maximum(sl, amp::abscomplex(lua(i,j)));
            }
            for(j=i; j<=n-1; j++)
            {
                su = amp::maximum(su, amp::abscomplex(lua(i,j)));
            }
        }
        if( su==0 )
        {
            su = 1;
        }
        su = 1/su;
        sl = 1/sl;
        
        //
        // Estimate the norm of SU*SL*A.
        //
        if( !isanormprovided )
        {
            anorm = 0;
            if( onenorm )
            {
                kase1 = 1;
            }
            else
            {
                kase1 = 2;
            }
            kase = 0;
            do
            {
                cmatrixestimatenorm(n, cwork4, ex, anorm, kase, isave, rsave);
                if( kase!=0 )
                {
                    if( kase==kase1 )
                    {
                        
                        //
                        // Multiply by U
                        //
                        for(i=1; i<=n; i++)
                        {
                            i1_ = (i)-(i-1);
                            v = 0.0;
                            for(i_=i-1; i_<=n-1;i_++)
                            {
                                v += lua(i-1,i_)*ex(i_+i1_);
                            }
                            ex(i) = v;
                        }
                        
                        //
                        // Multiply by L
                        //
                        for(i=n; i>=1; i--)
                        {
                            v = 0;
                            if( i>1 )
                            {
                                i1_ = (1)-(0);
                                v = 0.0;
                                for(i_=0; i_<=i-2;i_++)
                                {
                                    v += lua(i-1,i_)*ex(i_+i1_);
                                }
                            }
                            ex(i) = v+ex(i);
                        }
                    }
                    else
                    {
                        
                        //
                        // Multiply by L'
                        //
                        for(i=1; i<=n; i++)
                        {
                            cwork2(i) = 0;
                        }
                        for(i=1; i<=n; i++)
                        {
                            v = ex(i);
                            if( i>1 )
                            {
                                i1_ = (0) - (1);
                                for(i_=1; i_<=i-1;i_++)
                                {
                                    cwork2(i_) = cwork2(i_) + v*amp::conj(lua(i-1,i_+i1_));
                                }
                            }
                            cwork2(i) = cwork2(i)+v;
                        }
                        
                        //
                        // Multiply by U'
                        //
                        for(i=1; i<=n; i++)
                        {
                            ex(i) = 0;
                        }
                        for(i=1; i<=n; i++)
                        {
                            v = cwork2(i);
                            i1_ = (i-1) - (i);
                            for(i_=i; i_<=n;i_++)
                            {
                                ex(i_) = ex(i_) + v*amp::conj(lua(i-1,i_+i1_));
                            }
                        }
                    }
                }
            }
            while( kase!=0 );
        }
        
        //
        // Scale according to SU/SL
        //
        anorm = anorm*su*sl;
        
        //
        // Quick return if possible
        //
        if( anorm==0 )
        {
            return;
        }
        
        //
        // Estimate the norm of inv(A).
        //
        ainvnm = 0;
        if( onenorm )
        {
            kase1 = 1;
        }
        else
        {
            kase1 = 2;
        }
        kase = 0;
        while( true )
        {
            cmatrixestimatenorm(n, cwork4, ex, ainvnm, kase, isave, rsave);
            if( kase==0 )
            {
                break;
            }
            
            //
            // From 1-based to 0-based
            //
            for(i=0; i<=n-1; i++)
            {
                ex(i) = ex(i+1);
            }
            
            //
            // multiply by inv(A) or inv(A')
            //
            if( kase==kase1 )
            {
                
                //
                // Multiply by inv(L).
                //
                if( !safesolve::cmatrixscaledtrsafesolve(lua, sl, n, ex, false, 0, true, maxgrowth) )
                {
                    rc = 0;
                    return;
                }
                
                //
                // Multiply by inv(U).
                //
                if( !safesolve::cmatrixscaledtrsafesolve(lua, su, n, ex, true, 0, false, maxgrowth) )
                {
                    rc = 0;
                    return;
                }
            }
            else
            {
                
                //
                // Multiply by inv(U').
                //
                if( !safesolve::cmatrixscaledtrsafesolve(lua, su, n, ex, true, 2, false, maxgrowth) )
                {
                    rc = 0;
                    return;
                }
                
                //
                // Multiply by inv(L').
                //
                if( !safesolve::cmatrixscaledtrsafesolve(lua, sl, n, ex, false, 2, true, maxgrowth) )
                {
                    rc = 0;
                    return;
                }
            }
            
            //
            // from 0-based to 1-based
            //
            for(i=n-1; i>=0; i--)
            {
                ex(i+1) = ex(i);
            }
        }
        
        //
        // Compute the estimate of the reciprocal condition number.
        //
        if( ainvnm!=0 )
        {
            rc = 1/ainvnm;
            rc = rc/anorm;
            if( rc<rcondthreshold() )
            {
                rc = 0;
            }
        }
    }


    /*************************************************************************
    Internal subroutine for matrix norm estimation

      -- LAPACK auxiliary routine (version 3.0) --
         Univ. of Tennessee, Univ. of California Berkeley, NAG Ltd.,
         Courant Institute, Argonne National Lab, and Rice University
         February 29, 1992
    *************************************************************************/
    
    void rmatrixestimatenorm(int n,
        ap::template_1d_array< amp::ampf >& v,
        ap::template_1d_array< amp::ampf >& x,
        ap::template_1d_array< int >& isgn,
        amp::ampf& est,
        int& kase)
    {
        int itmax;
        int i;
        amp::ampf t;
        bool flg;
        int positer;
        int posj;
        int posjlast;
        int posjump;
        int posaltsgn;
        int posestold;
        int postemp;


        itmax = 5;
        posaltsgn = n+1;
        posestold = n+2;
        postemp = n+3;
        positer = n+1;
        posj = n+2;
        posjlast = n+3;
        posjump = n+4;
        if( kase==0 )
        {
            v.setlength(n+4);
            x.setlength(n+1);
            isgn.setlength(n+5);
            t = amp::ampf(1)/amp::ampf(n);
            for(i=1; i<=n; i++)
            {
                x(i) = t;
            }
            kase = 1;
            isgn(posjump) = 1;
            return;
        }
        
        //
        //     ................ ENTRY   (JUMP = 1)
        //     FIRST ITERATION.  X HAS BEEN OVERWRITTEN BY A*X.
        //
        if( isgn(posjump)==1 )
        {
            if( n==1 )
            {
                v(1) = x(1);
                est = amp::abs(v(1));
                kase = 0;
                return;
            }
            est = 0;
            for(i=1; i<=n; i++)
            {
                est = est+amp::abs(x(i));
            }
            for(i=1; i<=n; i++)
            {
                if( x(i)>=0 )
                {
                    x(i) = 1;
                }
                else
                {
                    x(i) = -1;
                }
                isgn(i) = amp::sign(x(i));
            }
            kase = 2;
            isgn(posjump) = 2;
            return;
        }
        
        //
        //     ................ ENTRY   (JUMP = 2)
        //     FIRST ITERATION.  X HAS BEEN OVERWRITTEN BY TRANDPOSE(A)*X.
        //
        if( isgn(posjump)==2 )
        {
            isgn(posj) = 1;
            for(i=2; i<=n; i++)
            {
                if( amp::abs(x(i))>amp::abs(x(isgn(posj))) )
                {
                    isgn(posj) = i;
                }
            }
            isgn(positer) = 2;
            
            //
            // MAIN LOOP - ITERATIONS 2,3,...,ITMAX.
            //
            for(i=1; i<=n; i++)
            {
                x(i) = 0;
            }
            x(isgn(posj)) = 1;
            kase = 1;
            isgn(posjump) = 3;
            return;
        }
        
        //
        //     ................ ENTRY   (JUMP = 3)
        //     X HAS BEEN OVERWRITTEN BY A*X.
        //
        if( isgn(posjump)==3 )
        {
            amp::vmove(v.getvector(1, n), x.getvector(1, n));
            v(posestold) = est;
            est = 0;
            for(i=1; i<=n; i++)
            {
                est = est+amp::abs(v(i));
            }
            flg = false;
            for(i=1; i<=n; i++)
            {
                if( x(i)>=0 && isgn(i)<0 || x(i)<0 && isgn(i)>=0 )
                {
                    flg = true;
                }
            }
            
            //
            // REPEATED SIGN VECTOR DETECTED, HENCE ALGORITHM HAS CONVERGED.
            // OR MAY BE CYCLING.
            //
            if( !flg || est<=v(posestold) )
            {
                v(posaltsgn) = 1;
                for(i=1; i<=n; i++)
                {
                    x(i) = v(posaltsgn)*(1+(amp::ampf(i-1))/(amp::ampf(n-1)));
                    v(posaltsgn) = -v(posaltsgn);
                }
                kase = 1;
                isgn(posjump) = 5;
                return;
            }
            for(i=1; i<=n; i++)
            {
                if( x(i)>=0 )
                {
                    x(i) = 1;
                    isgn(i) = 1;
                }
                else
                {
                    x(i) = -1;
                    isgn(i) = -1;
                }
            }
            kase = 2;
            isgn(posjump) = 4;
            return;
        }
        
        //
        //     ................ ENTRY   (JUMP = 4)
        //     X HAS BEEN OVERWRITTEN BY TRANDPOSE(A)*X.
        //
        if( isgn(posjump)==4 )
        {
            isgn(posjlast) = isgn(posj);
            isgn(posj) = 1;
            for(i=2; i<=n; i++)
            {
                if( amp::abs(x(i))>amp::abs(x(isgn(posj))) )
                {
                    isgn(posj) = i;
                }
            }
            if( x(isgn(posjlast))!=amp::abs(x(isgn(posj))) && isgn(positer)<itmax )
            {
                isgn(positer) = isgn(positer)+1;
                for(i=1; i<=n; i++)
                {
                    x(i) = 0;
                }
                x(isgn(posj)) = 1;
                kase = 1;
                isgn(posjump) = 3;
                return;
            }
            
            //
            // ITERATION COMPLETE.  FINAL STAGE.
            //
            v(posaltsgn) = 1;
            for(i=1; i<=n; i++)
            {
                x(i) = v(posaltsgn)*(1+(amp::ampf(i-1))/(amp::ampf(n-1)));
                v(posaltsgn) = -v(posaltsgn);
            }
            kase = 1;
            isgn(posjump) = 5;
            return;
        }
        
        //
        //     ................ ENTRY   (JUMP = 5)
        //     X HAS BEEN OVERWRITTEN BY A*X.
        //
        if( isgn(posjump)==5 )
        {
            v(postemp) = 0;
            for(i=1; i<=n; i++)
            {
                v(postemp) = v(postemp)+amp::abs(x(i));
            }
            v(postemp) = 2*v(postemp)/(3*n);
            if( v(postemp)>est )
            {
                amp::vmove(v.getvector(1, n), x.getvector(1, n));
                est = v(postemp);
            }
            kase = 0;
            return;
        }
    }


    
    void cmatrixestimatenorm(const int& n,
        ap::template_1d_array< amp::campf >& v,
        ap::template_1d_array< amp::campf >& x,
        amp::ampf& est,
        int& kase,
        ap::template_1d_array< int >& isave,
        ap::template_1d_array< amp::ampf >& rsave)
    {
        int itmax;
        int i;
        int iter;
        int j;
        int jlast;
        int jump;
        amp::ampf absxi;
        amp::ampf altsgn;
        amp::ampf estold;
        amp::ampf safmin;
        amp::ampf temp;
        int i_;


        
        //
        //Executable Statements ..
        //
        itmax = 5;
        safmin = amp::ampf::getAlgoPascalMinNumber();
        if( kase==0 )
        {
            v.setlength(n+1);
            x.setlength(n+1);
            isave.setlength(5);
            rsave.setlength(4);
            for(i=1; i<=n; i++)
            {
                x(i) = amp::ampf(1)/amp::ampf(n);
            }
            kase = 1;
            jump = 1;
            internalcomplexrcondsaveall(isave, rsave, i, iter, j, jlast, jump, absxi, altsgn, estold, temp);
            return;
        }
        internalcomplexrcondloadall(isave, rsave, i, iter, j, jlast, jump, absxi, altsgn, estold, temp);
        
        //
        // ENTRY   (JUMP = 1)
        // FIRST ITERATION.  X HAS BEEN OVERWRITTEN BY A*X.
        //
        if( jump==1 )
        {
            if( n==1 )
            {
                v(1) = x(1);
                est = amp::abscomplex(v(1));
                kase = 0;
                internalcomplexrcondsaveall(isave, rsave, i, iter, j, jlast, jump, absxi, altsgn, estold, temp);
                return;
            }
            est = internalcomplexrcondscsum1(x, n);
            for(i=1; i<=n; i++)
            {
                absxi = amp::abscomplex(x(i));
                if( absxi>safmin )
                {
                    x(i) = x(i)/absxi;
                }
                else
                {
                    x(i) = 1;
                }
            }
            kase = 2;
            jump = 2;
            internalcomplexrcondsaveall(isave, rsave, i, iter, j, jlast, jump, absxi, altsgn, estold, temp);
            return;
        }
        
        //
        // ENTRY   (JUMP = 2)
        // FIRST ITERATION.  X HAS BEEN OVERWRITTEN BY CTRANS(A)*X.
        //
        if( jump==2 )
        {
            j = internalcomplexrcondicmax1(x, n);
            iter = 2;
            
            //
            // MAIN LOOP - ITERATIONS 2,3,...,ITMAX.
            //
            for(i=1; i<=n; i++)
            {
                x(i) = 0;
            }
            x(j) = 1;
            kase = 1;
            jump = 3;
            internalcomplexrcondsaveall(isave, rsave, i, iter, j, jlast, jump, absxi, altsgn, estold, temp);
            return;
        }
        
        //
        // ENTRY   (JUMP = 3)
        // X HAS BEEN OVERWRITTEN BY A*X.
        //
        if( jump==3 )
        {
            for(i_=1; i_<=n;i_++)
            {
                v(i_) = x(i_);
            }
            estold = est;
            est = internalcomplexrcondscsum1(v, n);
            
            //
            // TEST FOR CYCLING.
            //
            if( est<=estold )
            {
                
                //
                // ITERATION COMPLETE.  FINAL STAGE.
                //
                altsgn = 1;
                for(i=1; i<=n; i++)
                {
                    x(i) = altsgn*(1+(amp::ampf(i-1))/(amp::ampf(n-1)));
                    altsgn = -altsgn;
                }
                kase = 1;
                jump = 5;
                internalcomplexrcondsaveall(isave, rsave, i, iter, j, jlast, jump, absxi, altsgn, estold, temp);
                return;
            }
            for(i=1; i<=n; i++)
            {
                absxi = amp::abscomplex(x(i));
                if( absxi>safmin )
                {
                    x(i) = x(i)/absxi;
                }
                else
                {
                    x(i) = 1;
                }
            }
            kase = 2;
            jump = 4;
            internalcomplexrcondsaveall(isave, rsave, i, iter, j, jlast, jump, absxi, altsgn, estold, temp);
            return;
        }
        
        //
        // ENTRY   (JUMP = 4)
        // X HAS BEEN OVERWRITTEN BY CTRANS(A)*X.
        //
        if( jump==4 )
        {
            jlast = j;
            j = internalcomplexrcondicmax1(x, n);
            if( amp::abscomplex(x(jlast))!=amp::abscomplex(x(j)) && iter<itmax )
            {
                iter = iter+1;
                
                //
                // MAIN LOOP - ITERATIONS 2,3,...,ITMAX.
                //
                for(i=1; i<=n; i++)
                {
                    x(i) = 0;
                }
                x(j) = 1;
                kase = 1;
                jump = 3;
                internalcomplexrcondsaveall(isave, rsave, i, iter, j, jlast, jump, absxi, altsgn, estold, temp);
                return;
            }
            
            //
            // ITERATION COMPLETE.  FINAL STAGE.
            //
            altsgn = 1;
            for(i=1; i<=n; i++)
            {
                x(i) = altsgn*(1+(amp::ampf(i-1))/(amp::ampf(n-1)));
                altsgn = -altsgn;
            }
            kase = 1;
            jump = 5;
            internalcomplexrcondsaveall(isave, rsave, i, iter, j, jlast, jump, absxi, altsgn, estold, temp);
            return;
        }
        
        //
        // ENTRY   (JUMP = 5)
        // X HAS BEEN OVERWRITTEN BY A*X.
        //
        if( jump==5 )
        {
            temp = 2*(internalcomplexrcondscsum1(x, n)/(3*n));
            if( temp>est )
            {
                for(i_=1; i_<=n;i_++)
                {
                    v(i_) = x(i_);
                }
                est = temp;
            }
            kase = 0;
            internalcomplexrcondsaveall(isave, rsave, i, iter, j, jlast, jump, absxi, altsgn, estold, temp);
            return;
        }
    }


    
    amp::ampf internalcomplexrcondscsum1(const ap::template_1d_array< amp::campf >& x,
        int n)
    {
        amp::ampf result;
        int i;


        result = 0;
        for(i=1; i<=n; i++)
        {
            result = result+amp::abscomplex(x(i));
        }
        return result;
    }


    
    int internalcomplexrcondicmax1(const ap::template_1d_array< amp::campf >& x,
        int n)
    {
        int result;
        int i;
        amp::ampf m;


        result = 1;
        m = amp::abscomplex(x(1));
        for(i=2; i<=n; i++)
        {
            if( amp::abscomplex(x(i))>m )
            {
                result = i;
                m = amp::abscomplex(x(i));
            }
        }
        return result;
    }


    
    void internalcomplexrcondsaveall(ap::template_1d_array< int >& isave,
        ap::template_1d_array< amp::ampf >& rsave,
        int& i,
        int& iter,
        int& j,
        int& jlast,
        int& jump,
        amp::ampf& absxi,
        amp::ampf& altsgn,
        amp::ampf& estold,
        amp::ampf& temp)
    {
        isave(0) = i;
        isave(1) = iter;
        isave(2) = j;
        isave(3) = jlast;
        isave(4) = jump;
        rsave(0) = absxi;
        rsave(1) = altsgn;
        rsave(2) = estold;
        rsave(3) = temp;
    }


    
    void internalcomplexrcondloadall(ap::template_1d_array< int >& isave,
        ap::template_1d_array< amp::ampf >& rsave,
        int& i,
        int& iter,
        int& j,
        int& jlast,
        int& jump,
        amp::ampf& absxi,
        amp::ampf& altsgn,
        amp::ampf& estold,
        amp::ampf& temp)
    {
        i = isave(0);
        iter = isave(1);
        j = isave(2);
        jlast = isave(3);
        jump = isave(4);
        absxi = rsave(0);
        altsgn = rsave(1);
        estold = rsave(2);
        temp = rsave(3);
    }
} // namespace

#endif
