//*********************************************************************************
//
//     This program is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 2 of the License, or
//     (at your option) any later version.
//
//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License along
//    with this program; if not, write to the Free Software Foundation, Inc.,
//    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//    
//    A copy of the GNU General Public License is also available at
//    http://www.fsf.org/licensing/licenses
//
//*********************************************************************************

void append_dw(gslgmp_vector* OUT, gslgmp_vector* IN, gslgmp_vector* OMEGA, int NOMEGA, int HPNUM, int OMEGAHPMIN, int OMEGAHPMAX)
{
	mpfr_t ptmp1,ptmp2,ptmp3;

	mpfr_init2(ptmp1,PRECBITS);
	mpfr_init2(ptmp2,PRECBITS);
	mpfr_init2(ptmp3,PRECBITS);

	if(HPNUM!=0){

		//*****************************************************************************************
		//	THE OMEGA INTERVAL IS DIVIDED INTO THREE PARTS
		//	[wmin,wmin^HP] [wmin^HP,wmax^HP] [wmax^HP,wmax]
		//	where the inner interval is discretized by dw^HP
		//	and the outer intevals with dw.
		//
		//	Any integral over omega now becomes:
		//
		//				 Sum_l=0^N-1     [ F(l) ]dw    + 1/2 F(wmin^HP) dw
		//	+ 1/2 F(wmin^HP) dw^HP + Sum_l=N+1^N'-1  [ F(l) ]dw^HP + 1/2 F(wmax^HP) dw^HP
		//	+ 1/2 F(wmax^HP) dw    + Sum_l=N'+1^NW-1 [ F(l) ]dw 
		//
		// 	To write this simply as a sum over an array this function multiplies each 
		//	component of the original function array F(l) with the appropriate dw term:
		//
		//	l < l(wmin^HP) 				fdw(l)= f(l) dw
		//	l = l(wmin^HP) 				fdw(l)= f(l) (1/2 dw + 1/2 dw^HP) 
		//	l(wmin^HP) < l < l(wmax^HP) 		fdw(l)= f(l) dw^HP
		//	l = l(wmax^HP) 				fdw(l)= f(l) (1/2 dw^HP + 1/2 dw) 
		//	l(wmax^HP) < l 				fdw(l)= f(l) dw	
		//*****************************************************************************************

		//*****************************************************************************************
		//	l < l(wmin^HP) 				fdw(l)= f(l) dw
		//*****************************************************************************************

		mpfr_sub(ptmp1,OMEGA->data[1],OMEGA->data[0],GMP_RNDN);

		for(int dw=0;dw<OMEGAHPMIN;dw++){
		mpfr_mul(OUT->data[dw],IN->data[dw],ptmp1,GMP_RNDN);
		}

		//*****************************************************************************************
		//	l = l(wmin^HP) 				fdw(l)= f(l) (1/2 dw + 1/2 dw^HP) 
		//*****************************************************************************************

		mpfr_sub(ptmp1,OMEGA->data[OMEGAHPMIN+1],OMEGA->data[OMEGAHPMIN],GMP_RNDN);
		mpfr_sub(ptmp2,OMEGA->data[OMEGAHPMIN],OMEGA->data[OMEGAHPMIN-1],GMP_RNDN);
		
		if(mpfr_equal_p(ptmp1,ptmp2)){ fprintf(stderr, "PROBLEM WITH THE HIGH PRECISION SUBINTERVAL DIVISION NOT CORRECT\n"); exit(-1);}
		
		mpfr_div_ui(ptmp1,ptmp1,2,GMP_RNDN);
		mpfr_div_ui(ptmp2,ptmp2,2,GMP_RNDN);
		mpfr_add(ptmp3,ptmp1,ptmp2,GMP_RNDN);
		mpfr_mul(OUT->data[OMEGAHPMIN],IN->data[OMEGAHPMIN],ptmp3,GMP_RNDN);

		
		//*****************************************************************************************
		//	l(wmin^HP) < l < l(wmax^HP) 		fdw(l)= f(l) dw^HP
		//*****************************************************************************************
		mpfr_sub(ptmp1,OMEGA->data[OMEGAHPMIN+2],OMEGA->data[OMEGAHPMIN+1],GMP_RNDN);
		
		for(int dw=OMEGAHPMIN+1;dw<OMEGAHPMAX;dw++){
		mpfr_mul(OUT->data[dw],IN->data[dw],ptmp1,GMP_RNDN);
		}


		//*****************************************************************************************
		//	l = l(wmax^HP) 				fdw(l)= f(l) (1/2 dw^HP + 1/2 dw) 
		//*****************************************************************************************
		
		mpfr_sub(ptmp1,OMEGA->data[OMEGAHPMAX+1],OMEGA->data[OMEGAHPMAX],GMP_RNDN);
		mpfr_sub(ptmp2,OMEGA->data[OMEGAHPMAX],OMEGA->data[OMEGAHPMAX-1],GMP_RNDN);
		
		mpfr_div_ui(ptmp1,ptmp1,2,GMP_RNDN);
		mpfr_div_ui(ptmp2,ptmp2,2,GMP_RNDN);
		mpfr_add(ptmp3,ptmp1,ptmp2,GMP_RNDN);
		mpfr_mul(OUT->data[OMEGAHPMAX],IN->data[OMEGAHPMAX],ptmp3,GMP_RNDN);

		//*****************************************************************************************
		//	l(wmax^HP) < l 				fdw(l)= f(l) dw	
		//*****************************************************************************************

		mpfr_sub(ptmp1,OMEGA->data[OMEGAHPMAX+2],OMEGA->data[OMEGAHPMAX+1],GMP_RNDN);

		for(int dw=OMEGAHPMAX+1;dw<NOMEGA;dw++){
		mpfr_mul(OUT->data[dw],IN->data[dw],ptmp1,GMP_RNDN);
		}

	}else{

		//*****************************************************************************************
		//	THE OMEGA INTERVAL IS EVENLY DISCRETIZED:
		//
		//	0 <= l < NOMEGA				fdw(l)= f(l) dw	
		//
		//*****************************************************************************************
		
		mpfr_sub(ptmp1,OMEGA->data[1],OMEGA->data[0],GMP_RNDN);

		for(int dw=0;dw<NOMEGA;dw++){
		mpfr_mul(OUT->data[dw],IN->data[dw],ptmp1,GMP_RNDN);
		}

	}

	mpfr_clear(ptmp1);
	mpfr_clear(ptmp2);
	mpfr_clear(ptmp3);
	

}