//*********************************************************************************
//
//     This program is free software; you can redistribute it and/or modify
//     it under the terms of the GNU General Public License as published by
//     the Free Software Foundation; either version 2 of the License, or
//     (at your option) any later version.
//
//     This program is distributed in the hope that it will be useful,
//     but WITHOUT ANY WARRANTY; without even the implied warranty of
//     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//     GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License along
//    with this program; if not, write to the Free Software Foundation, Inc.,
//    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//    
//    A copy of the GNU General Public License is also available at
//    http://www.fsf.org/licensing/licenses
//
//*********************************************************************************

void TESTSUITE()
{

	mpfr_t ptmp1,ptmp2,ptmp3,ptmp4,ptmp5,ptmp6;
	
	mpfr_init2(ptmp1,PRECBITS);
	mpfr_init2(ptmp2,PRECBITS);
	mpfr_init2(ptmp3,PRECBITS);
	mpfr_init2(ptmp4,PRECBITS);
	mpfr_init2(ptmp5,PRECBITS);
	mpfr_init2(ptmp6,PRECBITS);	
	
	//*********************************************************************************
	//	IN THIS FILE WE COLLECT SUBROUTINES THAT WILL PERFORM TESTS
	//	USING THE DIFFERENT SUBSYSTEMS OF THE PROGRAM TO CHECK WHETHER
	//	BUGS REMAIN IN THE CODE.
	//*********************************************************************************
	
	//*********************************************************************************
	//	CONSTRUCT MATRICES FOR WHICH THE FOLLOWING ROUTINE REQUIRE AT LEAST 128 
	//	BIT PRECISION
	//*********************************************************************************

	int SIZE=20;
	int SIZE2=30;
	
	int STARTW=-10;
	int ENDW=10;
	int STARTT=0;
	int ENDT=7;
	
	gslgmp_matrix *SYMMAT, *NSYMMAT, *NMMAT;
	gslgmp_matrix *EVEC, *S, *V, *INV;
	gslgmp_vector *EVAL, *REEVAL, *IMEVAL, *Xi;

	SYMMAT=gslgmp_matrix_calloc(SIZE,SIZE);
	NSYMMAT=gslgmp_matrix_calloc(SIZE,SIZE);
	NMMAT=gslgmp_matrix_calloc(SIZE,SIZE2);
	
	mpfr_set_ui(ptmp1, abs(ENDW-STARTW), GMP_RNDN);
	mpfr_div_ui(ptmp1,ptmp1,SIZE,GMP_RNDN);
	mpfr_set_ui(ptmp2, abs(ENDT-STARTT), GMP_RNDN);
	mpfr_div_ui(ptmp2,ptmp2,SIZE,GMP_RNDN);
	
	for(int i=0;i<SIZE; i++){
		for (int j=0; j<SIZE; j++){

			mpfr_set_si(ptmp3,STARTW,GMP_RNDN);
			mpfr_mul_ui(ptmp4,ptmp1,i,GMP_RNDN);
			mpfr_add(ptmp3,ptmp3,ptmp4,GMP_RNDN);
			
			mpfr_set_si(ptmp4,STARTW,GMP_RNDN);
			mpfr_mul_ui(ptmp5,ptmp1,j,GMP_RNDN);
			mpfr_add(ptmp4,ptmp4,ptmp5,GMP_RNDN);
			
			mpfr_mul(ptmp6,ptmp3,ptmp4,GMP_RNDN);
			mpfr_neg(ptmp6,ptmp6,GMP_RNDN);
			
			mpfr_exp(ptmp6,ptmp6,GMP_RNDN);
			
			gslgmp_matrix_set(SYMMAT,i,j,ptmp6);
		}
	}
	
	SYM_test(SYMMAT);
	//gmpPMat(SYMMAT);
	
	
	for(int i=0;i<SIZE; i++){
		for (int j=0; j<SIZE; j++){

			mpfr_set_si(ptmp3,STARTW,GMP_RNDN);
			mpfr_mul_ui(ptmp4,ptmp1,i,GMP_RNDN);
			mpfr_add(ptmp3,ptmp3,ptmp4,GMP_RNDN);
			
			mpfr_set_si(ptmp4,STARTT,GMP_RNDN);
			mpfr_mul_ui(ptmp5,ptmp2,j,GMP_RNDN);
			mpfr_add(ptmp4,ptmp4,ptmp5,GMP_RNDN);
			
			mpfr_mul(ptmp6,ptmp3,ptmp4,GMP_RNDN);
			mpfr_neg(ptmp6,ptmp6,GMP_RNDN);
			
			mpfr_exp(ptmp6,ptmp6,GMP_RNDN);
			
			gslgmp_matrix_set(NSYMMAT,i,j,ptmp6);
		}
	}

	SYM_test(NSYMMAT);
	//gmpPMat(NSYMMAT);
	
	for(int i=0;i<SIZE; i++){
		for (int j=0; j<SIZE2; j++){

			mpfr_set_si(ptmp3,STARTW,GMP_RNDN);
			mpfr_mul_ui(ptmp4,ptmp1,i,GMP_RNDN);
			mpfr_add(ptmp3,ptmp3,ptmp4,GMP_RNDN);
			
			mpfr_set_si(ptmp4,STARTT,GMP_RNDN);
			mpfr_mul_ui(ptmp5,ptmp2,j,GMP_RNDN);
			mpfr_add(ptmp4,ptmp4,ptmp5,GMP_RNDN);
			
			mpfr_mul(ptmp6,ptmp3,ptmp4,GMP_RNDN);
			mpfr_neg(ptmp6,ptmp6,GMP_RNDN);
			
			mpfr_exp(ptmp6,ptmp6,GMP_RNDN);
			
			gslgmp_matrix_set(NMMAT,i,j,ptmp6);
		}
	}

	//gmpPMat(NMMAT);


	//*********************************************************************************
	//	SVD DECOMPOSITION
	//*********************************************************************************

	int NSING=0;

	S=gslgmp_matrix_calloc(SIZE2,SIZE);
	V=gslgmp_matrix_calloc(SIZE,SIZE);
	Xi=gslgmp_vector_calloc(SIZE);
	
	gmpSVD_Transpose(NMMAT, S, V, Xi, &NSING);			
	
	
	SVDTrans_test(NMMAT,S,V,Xi);


	//*********************************************************************************
	//	SYMMETRIC EIGENSYSTEM
	//*********************************************************************************

	EVEC=gslgmp_matrix_calloc(SIZE,SIZE);
	EVAL=gslgmp_vector_calloc(SIZE);
	
	gmpEigenSYS_SYMM(SYMMAT,EVEC, EVAL);

	double TSTDEV;
	
	EiV_test(SYMMAT,EVEC,EVAL,&TSTDEV);

	gmpPVec(EVAL);
	
	//*********************************************************************************
	//	QR INVERTER
	//*********************************************************************************
	
	INV=gslgmp_matrix_calloc(SIZE,SIZE);
	
	gmpQRInvert(NSYMMAT,INV);
	INV_test(NSYMMAT,INV);


	//*********************************************************************************
	//	NON-SYMMETRIC EIGENVALUES
	//*********************************************************************************

	REEVAL=gslgmp_vector_calloc(SIZE);
	IMEVAL=gslgmp_vector_calloc(SIZE);
	
	Precision=PRECBITS;
	
	ap::template_1d_array< amp::ampf > ALG_Evalr;
	ap::template_1d_array< amp::ampf > ALG_Evali;
	
	ap::template_2d_array< amp::ampf > ALG_Evecr;
	ap::template_2d_array< amp::ampf > ALG_Eveci;

	ap::template_2d_array< amp::ampf > ALG_Lambda;
//fprintf(stderr,"DEBUG3\n");

	ALG_Evalr.setlength(SIZE);
	ALG_Evali.setlength(SIZE);
	ALG_Lambda.setlength(SIZE,SIZE);
	ALG_Evecr.setlength(SIZE,SIZE);
	ALG_Eveci.setlength(SIZE,SIZE);
//fprintf(stderr,"DEBUG4\n");

	for (int i=0;i<SIZE;i++)
	for (int j=0;j<SIZE;j++){
		
		gslgmp_matrix_get(SYMMAT,i,j,ptmp1);
		mpfr_set(ALG_Lambda(i,j).getWritePtr(),ptmp1,GMP_RNDN);
		
	}
//fprintf(stderr,"DEBUG5\n");

	int Alen=SIZE;
	int Avecs=0;

	evd::rmatrixevd(ALG_Lambda, Alen, Avecs, ALG_Evalr, ALG_Evali, ALG_Evecr, ALG_Eveci);

	for (int j=0;j<SIZE;j++){
		
		mpfr_set(REEVAL->data[j],ALG_Evalr(j).getReadPtr(),GMP_RNDN);
		mpfr_set(IMEVAL->data[j],ALG_Evali(j).getReadPtr(),GMP_RNDN);
		
	}

	gmpPVec(REEVAL);
	gmpPVec(IMEVAL);

	exit(-1);
}